"""Provides a function for creating mockmongo query expressions from HEA REST API parameters.
"""
from heaobject import user, group, root
from aiohttp.web import Request
from typing import Any
from collections.abc import Sequence, Mapping
import logging
from copy import deepcopy


def mongo_expr(request: Request | None, var_parts: str | Sequence[str] | None,
               mongoattributes: str | Sequence[str] | Mapping[str, Any] | None = None,
               extra: Mapping[str, Any] | None = None) -> dict[str, Any]:
    """
    Create and returns a mockmongo query expression representing filter criteria.

    1. If mongoattributes is a string, then mongoattributes is treated as a mockmongo field name, and var_parts is
    expected to be the name of a variable in the request's match_info with the desired value of the field name.

    2. If var_parts is a string and mongoattributes is not specified or None, the value of var_parts is treated as a
    mongo field name, and it is also the name of a variable in the request's match_info with the desired value of the
    field name.

    3. If mongoattributes is a Mapping, then it is treated as a mockmongo query expression, and the value of var_parts
    is ignored.

    4. If mongoattributes is a Sequence of strings, then it is treated as an array of mockmongo field names, and var_parts
    is expected to be a Sequence of aiohttp match_info strings with the desired values of the corresponding field names
    in mongoattributes.

    If var_parts is a Sequence of strings and mongoattributes is unspecified or None, then it is treated as both an
    iterable of mockmongo field names and corresponding aiohttp dynamic resource variable parts with the desired values of
    the corresponding field names.

    If mongoattributes or var_parts are specified but neither is a Mapping, a Sequence of strings nor a string, a
    TypeError will be raised.

    5. If mongoattributes and var_parts are both unspecified or None, an empty mockmongo query expression is created.

    If extra is a Mapping, it will be merged with the query expression dict above, overriding any overlapping parts of
    the expression. If extra is not None and is not a Mapping, a TypeError will be raised.

    The resulting query expression is returned.

    :param request: the aiohttp request. Required if var_parts is used.
    :param var_parts: the names of the dynamic resource's variable parts.
    :param mongoattributes: the attribute(s) to filter by, or a mockmongo query expression.
    :param extra: another mockmongo query expression.
    :return: a dict containing a mockmongo query expression.
    :raises TypeError: as described above.
    """
    logger = logging.getLogger(__name__)
    if isinstance(mongoattributes, str) and isinstance(var_parts, str):
        # 1 above
        if request is None:
            raise TypeError('request must be not None when var_parts is used')
        d: dict[str, Any] = {mongoattributes: request.match_info[var_parts]}
    elif not mongoattributes and isinstance(var_parts, str):
        # 2 above
        if request is None:
            raise TypeError('request must be not None when var_parts is used')
        d = {var_parts: request.match_info[var_parts]}
    elif isinstance(mongoattributes, Mapping):
        # 3 above
        d = dict(deepcopy(mongoattributes))
    elif mongoattributes or var_parts:
        # 4 above
        if var_parts is None:
            raise ValueError('var_parts cannot be None in this situation')
        if isinstance(var_parts, str):
            raise TypeError('var_parts cannot be a str in this situation')
        if mongoattributes is not None and isinstance(mongoattributes, str):
            raise TypeError('mongoattributes cannot be a str in this situation')
        if request is None:
            raise ValueError('request must be not None when var_parts is used')
        d = {nm: request.match_info[var_parts[idx]]
             for idx, nm in enumerate(mongoattributes if mongoattributes else var_parts)}
    else:
        # 5 above
        d = {}
    if extra:
        extra_ = dict(deepcopy(extra))
        if '$or' in extra_ and '$or' in d:
            if '$and' not in d:
                d.update({'$and': [{'$or': extra_.pop('$or')}, {'$or': d.pop('$or')}]})
            else:
                d['$and'].append({'$or': extra_.pop('$or')})
        if '$and' in extra_ and '$and' in d:
            d['$and'].extend(extra_.pop('$and'))
        d.update(extra_)
    logger.debug('Mongo expression is %s', d)
    return d


async def sub_filter_expr(permissions: Sequence[str], context: root.PermissionContext | None = None) -> dict[str, Any] | None:
    """
    Returns mongodb expression that filters results by user and permissions. The expression contains the following
    filters: 1) the user is the object's owner; 2) the user has a share for the object either directly or through group
    membership, and the permissions for the shares overlap with the requested permissions; 3) the object grants super
    admins access, the user is a super admin, and the requested permissions overlap with the super admin default
    permissions.

    :param permissions: the requested permissions to filter by.
    :param context: the context for getting group info.
    :return: a dict, or None if no context was provided.
    """
    if context:
        sub = context.sub
        # System users cannot be members of groups, so save the potential network lookup.
        groups = await context.get_groups() if not user.is_system_user(sub) else []
        or_clause = [{'owner': sub},
                     {'dynamic_permission_supported': True},
                     {'shares': {
                         '$elemMatch': {
                            '$or': [{'user': {'$in': _matching_users(sub)}}, {'group': {'$in': groups}}], 'permissions': {'$elemMatch': {'$in': permissions}}  # type: ignore[dict-item]
                     }
                    }}]
        # Save the potential network lookup if there is no group membership to check.
        if groups and await context.group_id_from(group.SUPERADMIN_GROUP) in groups:
            or_clause.append({'super_admin_default_permissions': {'$exists': True, '$type': 'array', '$elemMatch': {'$in': permissions}}})
        return {'$or': or_clause}
    else:
        return None


def _matching_users(sub: str) -> list[str]:
    """
    Returns a list containing the provided user and generic system users.

    :param sub: the user.
    :return: a list of users.
    """
    return [sub, user.ALL_USERS]




from ..db.aws import S3Manager, S3WithMongo, S3
from .dockermongo import DockerMongoManager, MockDockerMongoManager
from .mockaws import MockS3Manager, MockS3WithMongo


class S3WithDockerMongoManager(S3Manager, DockerMongoManager):
    """Database manager for connecting to real AWS S3 and a MongoDB docker container. The registry and volume
    microservices need to be running."""

    def get_database(self) -> S3WithMongo:
        return S3WithMongo(config=self.config)

    @classmethod
    def database_types(self) -> list[str]:
        return ['system|awss3', 'system|mongo']


class MockS3WithMockDockerMongoManager(MockS3Manager, MockDockerMongoManager):
    """
    Database manager for mocking AWS S3 buckets with moto. Mark test fixture data that is managed in S3 buckets with
    this database manager in unit test environments. Furthermore, connections to boto3/moto clients normally require
    access to the registry and volume microservices. This database manager mocks those connections. Mark
    component, volume, and filesystem test collections with this database manager to make them available in unit
    testing environments. This class is not designed to be subclassed.
    """

    def get_database(self) -> MockS3WithMongo:
        return MockS3WithMongo(config=self.config)

    @classmethod
    def database_types(self) -> list[str]:
        return ['system|awss3', 'system|mongo']



"""
Manages background tasks.
"""
from asyncio import Task, CancelledError, Queue, create_task, gather
from collections.abc import Coroutine
from typing import Any, Callable, Optional
from aiohttp import web
import logging

class BackgroundTasks:
    """
    Manages asyncio background tasks. An instance of this class is available as an application property with name
    HEA_BACKGROUND_TASKS.
    """
    def __init__(self, app: web.Application | None = None, queue_size= 0):
        """
        Creates a background task manager with the provided queue size.

        :param queue_size: the queue size. A value of zero means infinite size.
        """
        self.__app = app
        self.__tasks: dict[str, Task] = dict()
        self.__done: dict[str, Exception | None] = dict()
        self.__queue = Queue[Task](queue_size)
        self.__metadata: dict[str, dict[str, Any]] = dict()
        self.__result: dict[str, Any] = dict()
        self.__logger = logging.getLogger(__name__)

    def contains(self, name: str) -> bool:
        """
        Whether this background task manager is managing a task with the given name.

        :param name: the name of the background task.
        :return True or False.
        """
        return name in self.__tasks or name in self.__done

    async def add(self, coro: Callable[[Optional[web.Application]], Coroutine[Any, Any, None]], name: str | None = None, **metadata) -> str:
        """
        A coroutine for creating a new asyncio task from the given coroutine and adds it to the tasks managed by this
        task manager. If a name is provided, the task will be given the provided name. Otherwise, a name will be
        generated automatically.

        :param coro: the coroutine (required).
        :param name: an optional name.
        :param metadata: any key-value pairs associated with the task.
        :return: the name of the created task.
        """
        self.__logger.debug('Adding coroutine %s...', name)
        task = create_task(coro(self.__app), name=name)
        name_ = task.get_name()
        self.__tasks[name_] = task
        self.__metadata[name_] = metadata
        await self.__queue.put(task)
        self.__logger.debug('Coroutine %s added.', name_)
        return name_

    def metadata(self, name: str, key: str) -> Any:
        """
        Retrieve task metadata.

        :param name: the name of the task.
        :param key: the key of the metadata value of interest.
        :return: the metadata value.
        :raises KeyError: if the given task has no metadata with the given key.
        """
        return self.__metadata[name][key]

    def cancel(self, name: str):
        """
        Requests cancellation of the task with the given name.

        :param name: the name of the task to cancel.
        """
        self.__tasks[name].cancel()

    def cancel_all(self):
        """
        Requests cancellation of all tasks managed by this background task manager.
        """
        for task in self.__tasks.values():
            task.cancel()

    async def join(self, name: str):
        """
        A coroutine for awaiting completion of the task with the given name.

        :param name: the name of the task.
        """
        try:
            await self.__tasks[name]
        except:
            pass
        finally:
            self._mark_done(name)

    async def join_all(self) -> None:
        """
        A coroutine for awaiting completion of all tasks managed by this backround task manager.
        """
        done_tasks: list[str] = []
        for name, task in self.__tasks.items():
            try:
                await task
            except:
                pass
            finally:
                done_tasks.append(name)
        for name in done_tasks:
            self._mark_done(name)

    def in_progress(self, name: str) -> bool:
        """
        Returns whether the task is managed by this task manager and is still running.

        :param name: the name of the task.
        :return: True or False.
        """
        return self.contains(name) and not self.done(name)

    def done(self, name: str) -> bool:
        """
        Returns whether the task with the given name is completed.

        :param name: the name of the task to check.
        :return: True or False.
        """
        return name in self.__done

    def succeeded(self, name: str) -> bool:
        """
        Returns whether the task with the given name completed successfully.

        :param name: the name of the task to check.
        :return: True or False.
        """
        return bool(self.done(name) and not self.error(name))

    def failed(self, name: str) -> bool:
        """
        Returns whether the task with the given name failed.

        :param name: the name of the task to check.
        :return: True or False.
        """
        return bool(self.done(name) and self.error(name))

    def error(self, name: str) -> Exception | None:
        """
        Returns the exception raised within a failed task.

        :param name: the name of the task to check.
        :return: An exception if the task failed, or None if it succeeded.
        :raises ValueError: if the task is not done or not managed by this background task manager.
        """
        if name not in self.__done:
            raise ValueError(f'{name} not done or not managed by this background task manager.')
        return self.__done[name]

    def result(self, name: str) -> Any:
        if name not in self.__done:
            raise ValueError(f'{name} not done or not managed by this background task manager.')
        return self.__result[name]

    def remove(self, name: str):
        """
        Removes the task with the given name from this task manager. This method does nothing if no task with the given
        name is managed by this background task manager.

        :param name: the name of the task to remove.
        """
        if name in self.__tasks:
            del self.__tasks[name]
        if name in self.__done:
            del self.__done[name]
        if name in self.__metadata:
            del self.__metadata[name]

    async def clear(self):
        """
        A coroutine that cancels all running tasks and cleans up all resources.
        """
        self.cancel_all()
        await self.join_all()
        while not self.__queue.empty():
            self.__queue.get_nowait()
            self.__queue.task_done()
        self.__tasks.clear()
        self.__done.clear()
        self.__metadata.clear()

    async def auto_join(self):
        """
        A coroutine that automatically executes tasks added to this task manager. It runs forever. When wrapped in a
        task, cancelling the task ends this coroutine.
        """
        async def consumer():
            while task := await self.__queue.get():
                try:
                    await self.join(task.get_name())
                    self.__queue.task_done()
                except CancelledError:
                    await self.join(task.get_name())
                    self.__queue.task_done()
                    raise


        consumers = [create_task(consumer()) for _ in range(5)]
        try:
            await gather(*consumers)
        except CancelledError:
            for c in consumers:
                c.cancel()
            raise



    def __len__(self) -> int:
        """
        Returns the number of tasks managed by this background task manager.

        :return: the number of tasks managed by this background task manager.
        """
        return len(self.__tasks) + len(self.__done)

    def running_tasks(self) -> int:
        """
        Returns the number of running tasks.

        :return: the number of running tasks.
        """
        return len(self.__tasks)

    def done_tasks(self) -> int:
        """
        Returns the number of done tasks.

        :return: the number of done tasks.
        """
        return len(self.__done)

    def _mark_done(self, name: str):
        """
        Marks the task with the given name as done.

        :param name: the name of the task to mark done.
        """
        try:
            self.__result[name] = self.__tasks[name].result()
            self.__done[name] = None
        except CancelledError:
            self.__done[name] = None
            self.__result[name] = None
        except Exception as e:
            self.__done[name] = e
        self.__tasks.pop(name, None)

    def __str__(self) -> str:
        return f'Background task manager: in-progress tasks: {", ".join(self.__tasks)}; done tasks: {", ".join(self.__done)}'



from .clienttestcase import ClientTestCase as ClientTestCase_, fixtures
from heaserver.service import client
from heaobject.registry import Component
from heaobject.organization import Organization


class ClientTestCase(ClientTestCase_):

    def setUp(self) -> None:
        super().setUp()
        self.expected: Component = Component()
        self.expected.from_dict(fixtures['components'][0])

    async def test_get_no_component(self):
        self.assertEqual(None, await client.get_component(self.app, Organization.get_type_name()))

    async def test_get_component(self):
        actual = await client.get_component(self.app, self.expected.resources[0].resource_type_name)
        self.assertEqual(self.expected, actual)

    async def test_get_no_resource_url(self):
        self.assertEqual(None, await client.get_resource_url(self.app, Organization.get_type_name()))

    async def test_get_resource_url(self):
        self.assertEqual('http://localhost/foo/folders', await client.get_resource_url(self.app, self.expected.resources[0].resource_type_name))



from .componentpermissionstestcase import ComponentPermissionsTestCase
from heaserver.service.testcase.mixin import PermissionsPostMixin, PermissionsPutMixin, PermissionsGetOneMixin, \
    PermissionsGetAllMixin, PermissionsDeleteMixin


class TestPostComponentWithBadPermissions(ComponentPermissionsTestCase, PermissionsPostMixin):
    """A test case class for testing POST requests with bad permissions."""
    pass


class TestPutComponentWithBadPermissions(ComponentPermissionsTestCase, PermissionsPutMixin):
    """A test case class for testing PUT requests with bad permissions."""
    pass


class TestGetOneComponentWithBadPermissions(ComponentPermissionsTestCase, PermissionsGetOneMixin):
    """A test case class for testing GET one requests with bad permissions."""
    pass


class TestGetAllComponentsWithBadPermissions(ComponentPermissionsTestCase, PermissionsGetAllMixin):
    """A test case class for testing GET all requests with bad permissions."""
    pass


class TestDeleteComponentsWithBadPermissions(ComponentPermissionsTestCase, PermissionsDeleteMixin):
    """A test case class for testing DELETE requests with bad permissions."""
    pass



"""
Creates a test case class for use with the unittest library that is built into Python.
"""

from heaserver.service.testcase.dockermongo import MockDockerMongoManager
from heaserver.service.testcase.microservicetestcase import get_test_case_cls_default
from heaserver.service.testcase import expectedvalues, TEST_USER
from . import service
from heaobject.user import NONE_USER
from heaobject.root import Permission, HEAObjectDict
from heaobject.volume import DEFAULT_FILE_SYSTEM
from typing import Dict, List
from datetime import datetime, timezone


fixtures: Dict[str, List[HEAObjectDict]] = {
    service.MONGODB_COMPONENT_COLLECTION: [
        # No permissions
        {
            'id': '666f6f2d6261722d71757578',
            'created': datetime(2022, 6, 13, 19, 49, 21, 534455, tzinfo=timezone.utc),
            'derived_by': None,
            'derived_from': ['foo', 'bar'],
            'description': None,
            'display_name': 'Luximus',
            'invites': [],
            'modified': datetime(2022, 6, 13, 19, 49, 21, 534455, tzinfo=timezone.utc),
            'name': 'luximus',
            'owner': NONE_USER,
            'shares': [],
            'source': None,
            'source_detail': None,
            'type': 'heaobject.registry.Component',
            'base_url': 'http://localhost',
            'resources': [{
                'type': 'heaobject.registry.Resource',
                'resource_type_name': 'heaobject.folder.Folder',
                'base_path': 'folders',
                'file_system_name': DEFAULT_FILE_SYSTEM,
                'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                'resource_collection_type_display_name': 'heaobject.folder.Folder',
                'display_in_system_menu': False,
                'display_in_user_menu': False,
                'collection_mime_type': 'application/x.collection'
            },
            {
                'type': 'heaobject.registry.Resource',
                'resource_type_name': 'heaobject.folder.Item',
                'base_path': 'items',
                'file_system_name': DEFAULT_FILE_SYSTEM,
                'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                'resource_collection_type_display_name': 'heaobject.folder.Item',
                'display_in_system_menu': False,
                'display_in_user_menu': False,
                'collection_mime_type': 'application/x.collection'
            }],
            'super_admin_default_permissions': [Permission.VIEWER.name, Permission.EDITOR.name],
            'dynamic_permission_supported': False
        },
        # Viewer permission only
        {
            'id': '0123456789ab0123456789ab',
            'created': datetime(2022, 6, 13, 19, 49, 21, 534455, tzinfo=timezone.utc),
            'derived_by': None,
            'derived_from': ['foo', 'bar'],
            'description': None,
            'display_name': 'Reximus',
            'invites': [],
            'modified': datetime(2022, 6, 13, 19, 49, 21, 534455, tzinfo=timezone.utc),
            'name': 'reximus',
            'owner': NONE_USER,
            'shares': [{
                'type': 'heaobject.root.ShareImpl',
                'invite': None,
                'user': TEST_USER,
                'permissions': [Permission.VIEWER.name]
            }],
            'user_shares': [{
                'type': 'heaobject.root.ShareImpl',
                'invite': None,
                'user': TEST_USER,
                'permissions': [Permission.VIEWER.name]
            }],
            'group_shares': [],
            'source': None,
            'source_detail': None,
            'type': 'heaobject.registry.Component',
            'base_url': 'http://localhost',
            'resources': [{
                'type': 'heaobject.registry.Resource',
                'resource_type_name': 'heaobject.folder.Folder',
                'base_path': 'folders',
                'file_system_name': DEFAULT_FILE_SYSTEM,
                'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                'resource_collection_type_display_name': 'heaobject.folder.Folder',
                'display_in_system_menu': False,
                'display_in_user_menu': False,
                'collection_mime_type': 'application/x.collection'
            },
                {
                    'type': 'heaobject.registry.Resource',
                    'resource_type_name': 'heaobject.folder.Item',
                    'base_path': 'items',
                    'file_system_name': DEFAULT_FILE_SYSTEM,
                    'file_system_type': 'heaobject.volume.MongoDBFileSystem',
                    'resource_collection_type_display_name': 'heaobject.folder.Item',
                    'display_in_system_menu': False,
                    'display_in_user_menu': False,
                    'collection_mime_type': 'application/x.collection'
                }
            ],
            'super_admin_default_permissions': [Permission.VIEWER.name, Permission.EDITOR.name],
            'dynamic_permission_supported': False
        }
    ]
}

content = {
    service.MONGODB_COMPONENT_COLLECTION: {
        '666f6f2d6261722d71757578': b'The quick brown fox jumps over the lazy dog',
        '0123456789ab0123456789ab': b''
    }
}


ComponentPermissionsTestCase = \
    get_test_case_cls_default(coll=service.MONGODB_COMPONENT_COLLECTION, fixtures=fixtures,
                              duplicate_action_name='component-duplicate-form',
                              db_manager_cls=MockDockerMongoManager,
                              wstl_package=service.__package__, content=content, content_type='text/plain',
                              put_content_status=204,
                              href='http://localhost:8080/components/',
                              get_actions=[expectedvalues.Action(name='component-get-properties',
                                                                 rel=['hea-properties']),
                                           expectedvalues.Action(name='component-get-open-choices',
                                                                 url='http://localhost:8080/components/{id}/opener',
                                                                 rel=['hea-opener-choices']),
                                           expectedvalues.Action(name='component-duplicate',
                                                                 url='http://localhost:8080/components/{id}/duplicator',
                                                                 rel=['hea-duplicator'])],
                              get_all_actions=[
                                  expectedvalues.Action(name='component-get-properties',
                                                        rel=['hea-properties']),
                                  expectedvalues.Action(name='component-get-open-choices',
                                                        url='http://localhost:8080/components/{id}/opener',
                                                        rel=['hea-opener-choices']),
                                  expectedvalues.Action(name='component-duplicate',
                                                        url='http://localhost:8080/components/{id}/duplicator',
                                                        rel=['hea-duplicator'])],
                              expected_opener=expectedvalues.Link(
                                  url=f'http://localhost:8080/components/{fixtures[service.MONGODB_COMPONENT_COLLECTION][0]["id"]}/content',
                                  rel=['hea-default', 'hea-opener', 'text/plain']), sub=TEST_USER)



"""
Test HEA service.
"""
from heaobject.account import AWSAccount
from heaobject.organization import Organization

from heaserver.service import response, appproperty
from heaserver.service.runner import routes, web
from heaserver.service.db import mongoservicelib
from heaserver.service.wstl import action
from heaobject.registry import Component
from heaobject.volume import DEFAULT_FILE_SYSTEM, MongoDBFileSystem

MONGODB_COMPONENT_COLLECTION = 'components'
MONGODB_ORGANIZATION_COLLECTION = 'organizations'


@routes.get('/components/{id}')
@action('component-get-properties', rel='hea-properties')
@action('component-get-open-choices', rel='hea-opener-choices', path='components/{id}/opener')  # only add if the object is openable
@action('component-duplicate', rel='hea-duplicator', path='components/{id}/duplicator')
async def get_component(request: web.Request) -> web.Response:
    """
    Gets the component with the specified id.
    :param request: the HTTP request.
    :return: the requested component or Not Found.
    """
    return await mongoservicelib.get(request, MONGODB_COMPONENT_COLLECTION)


@routes.get('/components/byname/{name}')
async def get_component_by_name(request: web.Request) -> web.Response:
    """
    Gets the component with the specified name.
    :param request: the HTTP request.
    :return: the requested component or Not Found.
    """
    return await mongoservicelib.get_by_name(request, MONGODB_COMPONENT_COLLECTION)


@routes.get('/components/bytype/{type}')
@routes.get('/components/bytype/{type}/byfilesystemtype/{filesystemtype}')
@routes.get('/components/bytype/{type}/byfilesystemtype/{filesystemtype}/byfilesystemname/{filesystemname}')
async def get_components_by_type(request: web.Request) -> web.Response:
    """
    Gets the component that serves resources of the specified HEA object type.
    :param request: the HTTP request.
    :return: the requested component or Not Found.
    """
    if 'filesystemname' in request.match_info:
        file_system_name = request.match_info['filesystemname']
    else:
        file_system_name = DEFAULT_FILE_SYSTEM
    if file_system_name == DEFAULT_FILE_SYSTEM:
        query_clause = {
            '$or': [{'file_system_name': {'$exists': False}}, {'file_system_name': {'$in': [file_system_name, None]}}]}
    else:
        query_clause = {'file_system_name': {'$eq': file_system_name}}
    if 'filesystemtype' in request.match_info:
        file_system_type = request.match_info['filesystemtype']
    else:
        file_system_type = MongoDBFileSystem.get_type_name()
    if file_system_type == MongoDBFileSystem.get_type_name():
        query_clause.update({'$or': [{'file_system_type': {'$exists': False}},
                                     {'file_system_type': {'$in': [file_system_type, None]}}]})
    else:
        query_clause.update({'file_system_type': {'$eq': file_system_type}})
    mongo_attributes = {'resources': {
        '$elemMatch': {
            'resource_type_name': {'$eq': request.match_info['type']},
            **query_clause
        }}}
    result = await request.app[appproperty.HEA_DB].get(request,
                                                       MONGODB_COMPONENT_COLLECTION,
                                                       mongoattributes=mongo_attributes)
    return await response.get(request, result)


@routes.get('/components')
@routes.get('/components/')
@action('component-get-properties', rel='hea-properties')
@action('component-get-open-choices', rel='hea-opener-choices', path='components/{id}/opener')  # only add if the object is openable
@action('component-duplicate', rel='hea-duplicator', path='components/{id}/duplicator')
async def get_all_components(request: web.Request) -> web.Response:
    """
    Gets all components.
    :param request: the HTTP request.
    :return: all components.
    """
    return await mongoservicelib.get_all(request, MONGODB_COMPONENT_COLLECTION)


@routes.get('/components/{id}/duplicator')
@action(name='component-duplicate-form')
async def get_component_duplicator(request: web.Request) -> web.Response:
    """
    Gets a form template for duplicating the requested component.

    :param request: the HTTP request. Required.
    :return: the requested form, or Not Found if the requested component was not found.
    """
    return await mongoservicelib.get(request, MONGODB_COMPONENT_COLLECTION)


@routes.post('/components/duplicator')
async def post_component_duplicator(request: web.Request) -> web.Response:
    """
    Posts the provided component for duplication.
    :param request: the HTTP request.
    :return: a Response object with a status of Created and the object's URI in the
    """
    return await mongoservicelib.post(request, MONGODB_COMPONENT_COLLECTION, Component)


@routes.get('/components/{id}/opener')
@action('component-open-default', rel='hea-default hea-opener text/plain', path='components/{id}/content')
async def get_component_opener(request: web.Request) -> web.Response:
    """
    Opens the component. We wouldn't ordinarily open a component because there's nothing to open.

    :param request: the HTTP request. Required.
    :return: the requested form, or Not Found if the requested component was not found.
    """
    return await mongoservicelib.opener(request, MONGODB_COMPONENT_COLLECTION)


@routes.get('/components/{id}/content')
async def get_component_content(request: web.Request) -> web.Response:
    """
    Gets the component's content. Registry components have no content; this is just for testing purposes.

    :param request: the HTTP request. Required.
    :return: the requested content.
    """
    return await mongoservicelib.get_content(request, MONGODB_COMPONENT_COLLECTION)


@routes.post('/components')
@routes.post('/components/')
async def post_component(request: web.Request) -> web.Response:
    """
    Posts the provided component.
    :param request: the HTTP request.
    :return: a Response object with a status of Created and the object's URI in the
    """
    return await mongoservicelib.post(request, MONGODB_COMPONENT_COLLECTION, Component)


@routes.put('/components/{id}')
async def put_component(request: web.Request) -> web.Response:
    """
    Updates the component with the specified id.
    :param request: the HTTP request.
    :return: a Response object with a status of No Content or Not Found.
    """
    return await mongoservicelib.put(request, MONGODB_COMPONENT_COLLECTION, Component)


@routes.put('/components/{id}/content')
async def put_component_content(request: web.Request) -> web.Response:
    """
    Updates the component's content. Registry components have no content; this is just for testing purposes. There is
    no corresponding POST call because the content's initial creation happens automatically when a POST call to
    /components/ is sent.

    :param request: the HTTP request. Required.
    :return: the requested content.
    """
    return await mongoservicelib.put_content(request, MONGODB_COMPONENT_COLLECTION, Component)


@routes.delete('/components/{id}')
async def delete_component(request: web.Request) -> web.Response:
    """
    Deletes the component with the specified id.
    :param request: the HTTP request.
    :return: No Content or Not Found.
    """
    return await mongoservicelib.delete(request, MONGODB_COMPONENT_COLLECTION)


@routes.get('/organizations/{id}')
@action('heaserver-organizations-organization-get-properties', rel='properties')
@action('heaserver-organizations-organization-get-open-choices', rel='hea-opener-choices',
        path='organizations/{id}/opener')
@action('heaserver-organizations-organization-duplicate', rel='duplicator', path='organizations/{id}/duplicator')
async def get_organization(request: web.Request) -> web.Response:
    """
    Gets the organization with the specified id.
    :param request: the HTTP request.
    :return: the requested organization or Not Found.
    ---
    summary: A specific organization.
    tags:
        - heaserver-organizations-get-organization
    parameters:
        - $ref: '#/components/parameters/id'
    responses:
      '200':
        $ref: '#/components/responses/200'
      '404':
        $ref: '#/components/responses/404'
    """
    return await mongoservicelib.get(request, MONGODB_ORGANIZATION_COLLECTION)


@routes.get('/organizations/byname/{name}')
async def get_organization_by_name(request: web.Request) -> web.Response:
    """
    Gets the organization with the specified id.
    :param request: the HTTP request.
    :return: the requested organization or Not Found.
    ---
    summary: A specific organization, by name.
    tags:
        - heaserver-organizations-get-organization-by-name
    parameters:
        - $ref: '#/components/parameters/name'
    responses:
      '200':
        $ref: '#/components/responses/200'
      '404':
        $ref: '#/components/responses/404'
    """
    return await mongoservicelib.get_by_name(request, MONGODB_ORGANIZATION_COLLECTION)


@routes.get('/organizations/{id}/duplicator')
@action(name='heaserver-organizations-organization-duplicate-form', path='organizations/{id}')
async def get_organization_duplicate_form(request: web.Request) -> web.Response:
    """
    Gets a form template for duplicating the requested organization.

    :param request: the HTTP request. Required.
    :return: the requested form, or Not Found if the requested organization was not found.
    """
    return await mongoservicelib.get(request, MONGODB_ORGANIZATION_COLLECTION)


@routes.get('/organizations/{id}/opener')
@action('heaserver-organizations-organization-open-aws-accounts', rel='hea-default hea-opener text/plain',
        path='organizations/{id}/content')
async def get_organization_opener(request: web.Request) -> web.Response:
    """

    :param request: the HTTP Request.
    :return: A Response object with a status of Multiple Choices or Not Found.
    ---
    summary: Organization opener choices
    tags:
        - heaserver-organizations-organization-get-open-choices
    parameters:
        - $ref: '#/components/parameters/id'
    responses:
      '300':
        $ref: '#/components/responses/300'
      '404':
        $ref: '#/components/responses/404'
    """
    return await mongoservicelib.opener(request, MONGODB_ORGANIZATION_COLLECTION)


@routes.get('/organizations')
@routes.get('/organizations/')
@action('heaserver-organizations-organization-get-properties', rel='properties')
@action('heaserver-organizations-organization-get-open-choices', rel='hea-opener-choices',
        path='organizations/{id}/opener')
@action('heaserver-organizations-organization-duplicate', rel='duplicator', path='organizations/{id}/duplicator')
async def get_all_organizations(request: web.Request) -> web.Response:
    """
    Gets all organizations.
    :param request: the HTTP request.
    :return: all organizations.
    ---
    summary: All organizations.
    tags:
        - heaserver-organizations-get-all-organizations
    responses:
      '200':
        $ref: '#/components/responses/200'
    """
    return await mongoservicelib.get_all(request, MONGODB_ORGANIZATION_COLLECTION)


@routes.get('/organizations/{id}/content')
async def get_organization_content(request: web.Request) -> web.Response:
    """
    Gets the component's content. Registry components have no content; this is just for testing purposes.

    :param request: the HTTP request. Required.
    :return: the requested content.
    """
    return await mongoservicelib.get_content(request, MONGODB_ORGANIZATION_COLLECTION)


@routes.put('/organizations/{id}')
async def put_organization(request: web.Request) -> web.Response:
    """
    Updates the organization with the specified id.
    :param request: the HTTP request.
    :return: a Response object with a status of No Content or Not Found.
    """
    return await mongoservicelib.put(request, MONGODB_ORGANIZATION_COLLECTION, Organization)


@routes.put('/organizations/{id}/content')
async def put_organization_content(request: web.Request) -> web.Response:
    """
    Updates the component's content. Registry components have no content; this is just for testing purposes. There is
    no corresponding POST call because the content's initial creation happens automatically when a POST call to
    /components/ is sent.

    :param request: the HTTP request. Required.
    :return: the requested content.
    """
    return await mongoservicelib.put_content(request, MONGODB_ORGANIZATION_COLLECTION, Organization)



"""
Implements connectivity to a RabbitMQ message broker.
"""
import queue
import sys
from functools import partial
from heaobject.root import DesktopObject, from_json, desktop_object_from_json, desktop_object_type_for_name
from heaserver.service import appproperty
from heaserver.service.config import Configuration
from pathlib import Path
from aiohttp.web import Application
from aio_pika import Message, ExchangeType, connect_robust, DeliveryMode
from aio_pika.robust_exchange import AbstractExchange
import asyncio
import aiodiskqueue
from typing import Optional, Any
from asyncio import CancelledError
from contextlib import asynccontextmanager
from collections.abc import AsyncIterator, AsyncGenerator, Callable, Awaitable
import logging
import os

_logger = partial(logging.getLogger, __name__)

DEFAULT_EXCHANGE = 'hea_desktop_objects'
DEFAULT_HOSTNAME = 'localhost'
DEFAULT_PORT = 5672
DEFAULT_USERNAME = 'guest'
DEFAULT_PASSWORD = 'guest'
DEFAULT_TOPICS = ('heaobject.root.DesktopObject',)
CONFIG_SECTION = 'MessageBroker'

async def subscribe(hostname=DEFAULT_HOSTNAME,
                    port=DEFAULT_PORT,
                    username=DEFAULT_USERNAME,
                    password=DEFAULT_PASSWORD,
                    exchange_name=DEFAULT_EXCHANGE,
                    topics=DEFAULT_TOPICS,
                    prefetch_count=1) -> AsyncGenerator[DesktopObject, None]:
    """
    Asynchronous generator that connects to a RabbitMQ message broker and allows subscribing to desktop objects that
    are published to an exchange.

    :param hostname: RabbitMQ's hostname (localhost by default).
    :param port: RabbitMQ's port (5672 by default).
    :param username: A username for connecting to RabbitMQ (guest by default).
    :param password: A password for connecting to RabbitMQ (guest by default).
    :param exchange_name: the exchange on which to wait for desktop objects (hea_desktop_objects by default)).
    :param topics: a list of one or more desktop object type names to wait for. All subtypes will also be awaited.
    Ensure that all subtypes of interest have been previously loaded into the python interpreter with an import
    statement, the importlib module, or some other method.
    :param prefetch_count: optimizes for performance (usually just leave the default value).
    :return: an async generator of DesktopObjects.
    """
    logger = _logger()

    async with await connect_robust(host=hostname, port=port, login=username, password=password) as connection:
        channel = await connection.channel()
        try:
            exchange = await channel.declare_exchange(exchange_name, ExchangeType.TOPIC)
            queue = await channel.declare_queue(exclusive=True)
            for topic in topics:
                await queue.bind(exchange, routing_key=topic)
                type_ = desktop_object_type_for_name(topic)
                await queue.bind(exchange, routing_key=type_.get_type_name())
                for subtype in type_.get_subclasses():
                    await queue.bind(exchange, routing_key=subtype.get_type_name())
            if prefetch_count:
                await channel.set_qos(prefetch_count=prefetch_count)

            async with queue.iterator() as queue_iter:
                logger.debug('Got queue iterator')
                async for message in queue_iter:
                    logger.debug('Got message from queue iterator')
                    async with message.process():
                        message_body = message.body
                        logger.debug('Received message body: %s', message_body)
                        hea_object = desktop_object_from_json(message_body)
                        assert isinstance(hea_object, tuple(desktop_object_type_for_name(tn) for tn in topics)), f'Expected type {topics} but received {type(hea_object)}'
                        yield hea_object
        finally:
            logger.debug('Closing the channel and connection...')
            await channel.close()


def subscriber_cleanup_context_factory(message_body_cb: Optional[Callable[[Application, DesktopObject], Awaitable[Any]]] = None,
                                       config: Optional[Configuration] = None,
                                       hostname=None,
                                       port=None,
                                       username=None,
                                       password=None,
                                       exchange_name=DEFAULT_EXCHANGE,
                                       topics=DEFAULT_TOPICS,
                                       prefetch_count=1) -> Callable[[Application], AsyncIterator[None]]:
    """
    Creates a cleanup context co-routine for establishing a connection to RabbitMQ as a subscriber. It tries getting
    connection information (hostname, port, username, and password) from the HEA config file, if found. Connection
    information in the config file may be overridden by passing not-None values into this function for the hostname,
    port, username, and password parameters. Hostname, port, username, and password values that remain None are set to
    their default values ('localhost', 5672, 'guest', and 'guest', respectively). Use the topics parameter to set
    one or more desktop object type names that this subscriber will listen for. Use # as a wildcard, for example,
    heaobject.volume.# for any desktop object in the heaobject.volume module. The
    aiohttp.web.app[appproperty.HEA_MESSAGE_BROKER_SUBSCRIBER] property is set to the initialized subscriber object,
    which may be used in an async for loop as follows, where message_body has type bytes:

    async for message_body in app[appproperty.HEA_MESSAGE_BROKER_SUBSCRIBER]:
        print(f'Got {message_body}')

    :param message_body_cb: a coroutine that is invoked when a message is read from the message broker with the
    aiohttp Application object and a DesktopObject as parameters. Raise an exception to signal that the desktop object
    should not be acknowledged as received from the message broker.
    :param config: a heaserver.service.config.Configuration object. The parsed configuration data should have a
    MessageBroker section with four properties:

            Hostname = the hostname of the message broker, localhost by default.
            Port: the message broker's port, 5672 by default.
            Username: a username for connecting to the message broker, guest by default.
            Password: a password for connecting to the message broker, guest by default.

    :param hostname: the hostname of the message broker, localhost by default.
    :param port: the message broker's port, 5672 by default.
    :param exchange_name: an optional MongoDB connection string that will override any database connection
    string in a provided config file.
    :param username: an optional username that will override any username in the connection string.
    :param password: an optional password that will override any password in the connection string.
    :param exchange_name: the exchange name (required).
    :param topics: a list of one or more desktop object type names that this subscriber will listen for. Use # as a
    wildcard, for example, heaobject.volume.# for any desktop object in the heaobject.volume module.
    :param prefetch_count: optimizes for performance (usually just leave the default value).
    """
    _hostname = None
    _port = None
    _username = None
    _password = None
    _prefetch_count = None
    config_data = config.parsed_config if config else None
    if config_data and CONFIG_SECTION in config_data:
        _section = config_data[CONFIG_SECTION]
        _hostname = _section.get('Hostname')
        _port = _section.getint('Port')
        _username = _section.get('Username')
        _password = _section.get('Password')
        _prefetch_count = _section.get('PrefetchCount')
    if hostname is not None:
        _hostname = hostname
    if port is not None:
        _port = port
    if username is not None:
        _username = username
    if password is not None:
        _password = password
    if exchange_name is None:
        exchange_name = DEFAULT_EXCHANGE
    if _prefetch_count is None:
        _prefetch_count = prefetch_count

    if _hostname is None:
        _hostname = DEFAULT_HOSTNAME
    if _port is None:
        _port = DEFAULT_PORT
    if _username is None:
        _username = DEFAULT_USERNAME
    if _password is None:
        _password = DEFAULT_PASSWORD

    async def subscriber_cleanup_context(app: Application) -> AsyncIterator[None]:
        logger = logging.getLogger(__name__)

        async def subscriber_loop():
            while True:
                try:
                    logger.debug('Subscribing to topics %s from the message broker', topics)
                    queue = subscribe(_hostname, _port, _username, _password, exchange_name, topics, _prefetch_count)
                    logger.debug('Awaiting desktop objects from the message broker...')
                    async for desktop_object in queue:
                        logger.debug('Received desktop object from the message broker: %s', desktop_object)
                        try:
                            await message_body_cb(app, desktop_object)
                        except CancelledError:
                            raise
                        except Exception:
                            logger.exception('Error in message_body_cb')
                except CancelledError:
                    logger.debug('Subscriber loop is cancelled')
                    raise
                except Exception:
                    logger.exception('Error in subscriber loop, retrying...')
                    await asyncio.sleep(5)  # Wait before retrying
                finally:
                    await queue.aclose()

        the_task = asyncio.create_task(subscriber_loop())

        yield

        logger.debug('Cancelling the subscriber loop...')
        the_task.cancel()
        await the_task


    return subscriber_cleanup_context


async def publisher(hostname=DEFAULT_HOSTNAME,
                    port=DEFAULT_PORT,
                    username=DEFAULT_USERNAME,
                    password=DEFAULT_PASSWORD,
                    exchange_name=DEFAULT_EXCHANGE) -> AsyncGenerator[None, DesktopObject]:
    """
    Co-routine that connects to a RabbitMQ message broker and allows publishing desktop objects without having to
    reconnect every time. Call this function followed by anext on the returned asynchronous generator to establish
    a connection to RabbitMQ. Call the publisher's asend method with a desktop object to publish, and call its aclose
    method when done.

    :param hostname: RabbitMQ's hostname (localhost by default).
    :param port: RabbitMQ's port (5672 by default).
    :param username: A username for connecting to RabbitMQ (guest by default).
    :param password: A password for connecting to RabbitMQ (guest by default).
    :param exchange_name: the exchange on which to publish the desktop object (hea_desktop_objects by default)).
    """
    logger = logging.getLogger(__name__)
    if hostname is None:
        hostname = DEFAULT_HOSTNAME
    if port is None:
        port = DEFAULT_PORT
    if username is None:
        username = DEFAULT_USERNAME
    if password is None:
        password = DEFAULT_PASSWORD
    if exchange_name is None:
        exchange_name = DEFAULT_EXCHANGE
    logger.debug('Connecting to message queue at hostname %s, port %s, username %s', hostname, port, username)

    while True:
        logger.debug('Connecting to message broker...')
        async with await connect_robust(host=hostname, port=port, login=username, password=password) as connection:
            logger.debug('Got connection...')
            channel = await connection.channel()
            logger.debug('Got channel...')
            try:
                exchange = await channel.declare_exchange(exchange_name, ExchangeType.TOPIC)
                logger.debug('Got exchange')
                while True:
                    logger.debug('Awaiting message to send...')
                    await _process_desktop_object(desktop_object=(yield), exchange=exchange)
            finally:
                logger.debug('Closing channel and connection')
                await channel.close()


def publisher_cleanup_context_factory(config: Configuration | None = None,
                                      hostname: str | None = None,
                                      port: int | None = None,
                                      username: str | None = None,
                                      password: str | None = None,
                                      exchange_name=DEFAULT_EXCHANGE,
                                      appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER,
                                      publish_queue_persistence_path: str | None = None,
                                      failed_cb: Awaitable[Exception] | None = None) -> Callable[
    [Application], AsyncIterator[None]]:
    """
    Creates a cleanup context co-routine for establishing a connection to RabbitMQ as a publisher. It tries getting
    connection information (hostname, port, username, and password) from the HEA config file, if found. Connection
    information in the config file may be overridden by passing not-None values into this function for the hostname,
    port, username, password, and publish_queue_persistence_path parameters. Hostname, port, username, and password
    values that remain None are set to their default values ('localhost', 5672, 'guest', and 'guest', respectively).
    The aiohttp.web.app[appproperty.HEA_MESSAGE_BROKER_PUBLISHER] property is set to the initialized publisher object,
    which may be used as follows, where desktop_object is a desktop object:

    await aiohttp.web.app[appproperty.HEA_MESSAGE_BROKER_PUBLISHER].asend(desktop_object)

    The sent desktop object is put on a queue for future publication. If PublishQueuePersistencePath is specified in
    the config object, then the queue is persisted to that file. Otherwise an in-memory queue is used, which will be
    lost across microservice restarts.

    :param config: a heaserver.service.config.Configuration object. The parsed configuration data should have a
    MessageBroker section with four properties:

            Hostname = the hostname of the message broker, localhost by default.
            Port: the message broker's port, 5672 by default.
            Username: a username for connecting to the message broker, guest by default.
            Password: a password for connecting to the message broker, guest by default.
            PublishQueuePersistencePath: the path to a sqlite database file for persisting the queue of messages to
            be published. If not specified, an in-memory queue is used.

    :param hostname: the optional hostname of the message broker, which will override any hostname in the config file
    (localhost by default).
    :param port: the optional port of the message broker, which will override any port in the config file (5672 by
    default).
    :param username: an optional username that will override any username in the config file (guest by default).
    :param password: an optional password that will override any password in the config file (guest by default).
    :param exchange_name: the exchange name (required).
    :param appproperty_: the app property to user for accessing the publisher. For microservices that only need
    one publisher, the default (appproperty.HEA_MESSAGE_BROKER_PUBLISHER) is fine, but define custom app properties
    using this parameter for microservices that need multiple publishers. To avoid clashes with built-in app
    properties defined by heaserver, avoid app properties starting with HEA_.
    :param publish_queue_persistence_path: the path to a sqlite database file for persisting the queue of messages to
    publish. If not specified, and PublishQueuePersistencePath is not in the config object, an in-memory queue is used.
    """
    _hostname: str | None = None
    _port: int | None = None
    _username: str | None = None
    _password: str | None = None
    _publish_queue_persistence_path: str | None = None
    config_data = config.parsed_config if config else None
    if config_data and 'DEFAULT' in config_data and 'MessageBrokerEnabled' in config_data['DEFAULT'] and not config_data['DEFAULT'].getboolean('MessageBrokerEnabled'):
        async def do_nothing(app: Application) -> AsyncGenerator[None, None]:
            _logger().warning('Message broker functionality is disabled')
            yield
        return do_nothing

    # Get connection information from the config object.
    if config_data and CONFIG_SECTION in config_data:
        _section = config_data[CONFIG_SECTION]
        _hostname = _section.get('Hostname')
        _port = _section.getint('Port')
        _username = _section.get('Username')
        _password = _section.get('Password')
        _publish_queue_persistence_path = _section.get('PublishQueuePersistencePath')

    # Override with the values passed into this function, if any.
    if hostname is not None:
        _hostname = hostname
    if port is not None:
        _port = port
    if username is not None:
        _username = username
    if password is not None:
        _password = password
    if publish_queue_persistence_path is not None:
        _publish_queue_persistence_path = publish_queue_persistence_path

    # Set to default values if still None.
    if exchange_name is None:
        exchange_name = DEFAULT_EXCHANGE
    if _hostname is None:
        _hostname = DEFAULT_HOSTNAME
    if _port is None:
        _port = DEFAULT_PORT
    if _username is None:
        _username = DEFAULT_USERNAME
    if _password is None:
        _password = DEFAULT_PASSWORD
    if appproperty_ is None:
        appproperty_ = appproperty.HEA_MESSAGE_BROKER_PUBLISHER

    @asynccontextmanager
    async def new_async_queue(generator: AsyncGenerator[Any, Any]) -> AsyncIterator[asyncio.Queue | aiodiskqueue.Queue]:
        """
        Creates a new asynchronous queue that asends items placed on the queue to the provided async generator. It
        assumes that the generator has been "primed" already by calling the built-in anext() on it. It also assumes
        that the caller will call the generator's aclose() method. Upon exiting the context manager, the queue's
        resources are cleaned up automatically. If PublishQueuePersistencePath is set in the config object or passed
        into publisher_cleanup_context_factory, the queue is persisted to that file. Otherwise an in-memory queue is
        used, which will be lost across microservice restarts.

        :param generator: an async generator (required).
        :returns a newly created queue.
        """
        logger = _logger()
        logger.debug('Creating a new queue for publishing messages to RabbitMQ')
        if _publish_queue_persistence_path is None:
            logger.debug('Queue is in-memory')
            queue: asyncio.Queue | aiodiskqueue.Queue = asyncio.Queue()
        else:
            logger.debug('Queue is persisted at %s', _publish_queue_persistence_path)
            queue = await aiodiskqueue.Queue.create(_publish_queue_persistence_path,
                                                    cls_storage_engine=aiodiskqueue.engines.SqliteEngine)
            queue._unfinished_tasks = queue.qsize()
        async def task_done():
            """
            Marks the task as done in the queue.
            """
            if isinstance(queue, aiodiskqueue.Queue):
                try:
                    await queue.task_done()
                except CancelledError:
                    await queue.task_done()
                    raise
            else:
                queue.task_done()
        async def queue_task():
            try:
                while True:
                    object_to_send = await queue.get()
                    try:
                        await generator.asend(object_to_send)
                    except:
                        try:
                            await task_done()
                        finally:
                            # Place it back on the queue to retry later. The queue size is unbounded, so this put call
                            # should always succeed without blocking.
                            await queue.put(object_to_send)
                        raise
            except Exception as e:
                await failed_cb(e) if failed_cb else None
            finally:
                # Leave persisted queue intact for future use.
                if not isinstance(queue, aiodiskqueue.Queue):
                    while not queue.empty():
                        queue.get_nowait()
                        queue.task_done()
        queue_task_ = asyncio.create_task(queue_task())
        yield queue
        await queue.join()
        queue_task_.cancel()
        await queue_task_

    async def publisher_cleanup_context(app: Application) -> AsyncIterator[None]:
        _publisher = publisher(_hostname, _port, _username, _password, exchange_name)
        try:
            await anext(_publisher)
            async with new_async_queue(_publisher) as queue:
                app[appproperty_] = queue
                yield
                return
        finally:
            await _publisher.aclose()

    return publisher_cleanup_context


async def publish_desktop_object(app: Application, desktop_object: DesktopObject,
                                 appproperty_=appproperty.HEA_MESSAGE_BROKER_PUBLISHER):
    """
    Places the desktop object on a queue for publishing it to the message broker, if the service is configured with a
    message broker publisher. If not, then this coroutine does nothing.

    :param app: the Application object (required).
    :param desktop_object: a DesktopObject (required).
    :param appproperty_: the application property for the message broker, defaults to
    appproperty.HEA_MESSAGE_BROKER_PUBLISHER.
    """
    if appproperty_ is None:
        appproperty_ = appproperty.HEA_MESSAGE_BROKER_PUBLISHER
    if appproperty_ in app:
        await app[appproperty_].put(desktop_object)


async def _process_desktop_object(desktop_object: DesktopObject, exchange: AbstractExchange):
    topic = desktop_object.get_type_name()
    message_body = desktop_object.to_json()
    message = Message(message_body.encode('utf-8'), delivery_mode=DeliveryMode.PERSISTENT)
    await exchange.publish(message, routing_key=topic)
    _logger().debug('Sent %r', message)



"""
This module implements a simple API for launching a swagger UI for trying out a HEA microservice's REST APIs.

This module assumes the testcontainers package is installed. Do not import it into environments where testcontainers
will not be available, for example, in any code that needs to run outside automated testing or the SwaggerUI interface.
"""

from aiohttp_swagger3 import SwaggerDocs, SwaggerUiSettings
from aiohttp_swagger3.handlers import application_json
from aiohttp_swagger3.swagger_info import SwaggerInfo
from aiohttp import web
from importlib.metadata import version
from typing import Dict, List, Tuple, Callable, Iterable, Optional, Type, Mapping
from .testenv import app_context, DockerContainerConfig
from ..wstl import RuntimeWeSTLDocumentBuilder
from ..runner import client_session_cleanup_ctx
from ..db.database import CollectionKey, MicroserviceDatabaseManager, validate_collection_keys
from .dockermongo import DockerMongoManager, RealRegistryContainerConfig
from ..db.database import convert_to_collection_keys, FixtureKeyTypes
from heaobject.root import HEAObjectDict
from ..openapi import download_openapi_spec
from .util import maybe_docker_is_not_running


def run(project_slug: str,
        desktop_objects: Mapping[CollectionKey | str, List[HEAObjectDict]],
        routes: Iterable[Tuple[Callable, str, Callable]],
        content: Mapping[CollectionKey | str, Mapping[str, bytes]] | None = None,
        wstl_builder_factory: Optional[Callable[[], RuntimeWeSTLDocumentBuilder]] = None,
        registry_docker_image: Optional[str] = None,
        other_docker_images: Optional[List[DockerContainerConfig]] = None,
        db_manager_cls: Optional[Type[MicroserviceDatabaseManager]] = DockerMongoManager):
    """
    Launches a swagger UI for trying out the given HEA APIs. It downloads the HEA OpenAPI spec from gitlab.com. It
    launches Docker containers for any other projects specified in the other_docker_images argument. It also
    launches two MongoDB databases in Docker containers, one for the project being tested, and the other for the other
    containers. It inserts the given HEA objects into both databases, creates HEA Server Registry entries for all
    containers, and makes the given routes available to query in swagger. All containers run on Docker's default bridge
    network.

    This function creates two MongoDB databases because the project being tested is not running in a Docker container,
    so it needs any other containers' external URLs (starting with localhost) rather than Docker's internal network's
    IP addresses. To simplify this function's arguments, we load the exact same data into both MongoDB databases,
    except for the registry components, which differ only in their base URLs and names. The latter is set to the
    component's base URL. One of the databases will contain localhost base URLs, and the other will contain Docker
    bridge network IP addresses. This function logs each MongoDB container's mongo connection string at the info level
    so that you can tell which container is which.

    :param project_slug: the Gitlab project slug of interest. Required.
    :param desktop_objects: a mapping of mongo collection -> list of HEA objects as dicts. Required.
    :param routes: a list of three-tuples containing the command, path and collable of each route of interest. The
    commands are one of: aiohttp.web.get, aiohttp.web.delete, aiohttp.web.post, aiohttp.web.put or aiohttp.web.view.
    :param content: a mapping of mongo collection -> mapping of id -> content in bytes. Optional if the service's
    objects are not supposed to have content.
    :param wstl_builder_factory: a zero-argument callable that will return a RuntimeWeSTLDocumentBuilder. Optional if
    this service has no actions. Typically, you will use the heaserver.service.wstl.get_builder_factory function to
    get a factory object.
    :param registry_docker_image: an heaserver-registry docker image in REPOSITORY:TAG format, that will be launched
    after the MongoDB container is live.
    :param other_docker_images: optional list of ContainerSpec objects, which specify additional HEA microservice
    Docker containers to start.
    :param db_manager_cls: a database manager class from the heaserver.server.db package for the microservice being
    tested, if database connectivity is needed. Sets the appproperty.HEA_DB application property to this object. If
    omitted or None, heaserver.service.db.mongo.MongoManager is used.
    :raises OSError: If an error occurred accessing the OpenAPI spec.
    """
    if other_docker_images and registry_docker_image is None:
        raise ValueError('registry_docker_image must not be None when other_docker_images is not empty')
    if db_manager_cls is not None:
        db_ = db_manager_cls
    else:
        db_ = DockerMongoManager
    with maybe_docker_is_not_running():
        with download_openapi_spec() as open_api_spec_file:
            validate_collection_keys(desktop_objects)
            if content is not None:
                validate_collection_keys(content)
            with app_context(db_manager_cls=db_,
                            desktop_objects=convert_to_collection_keys(desktop_objects),
                            content=convert_to_collection_keys(content) if content else None,
                            other_docker_images=other_docker_images,
                            registry_docker_image=RealRegistryContainerConfig(image=registry_docker_image) if registry_docker_image is not None else None,
                            wstl_builder_factory=wstl_builder_factory) as (app, _):
                _init_swagger_docs(app, open_api_spec_file, project_slug, routes)
                app.cleanup_ctx.append(client_session_cleanup_ctx)
                web.run_app(app)


async def application_octet_stream(request: web.Request) -> Tuple[bytes, bool]:
    """
    Media type handler for application/octet-stream data.

    :param request: aiohttp request (required).
    :return: a two-tuple with the data, and whether the returned data is "raw" (untransformed).
    """
    return await request.read(), True


async def text_plain(request: web.Request) -> tuple[bytes, bool]:
    """
    Media type handler for text/plain data.

    :param request: aiohttp request (required).
    :return: a two-tuple with the data, and whether the returned data is "raw" (untransformed).
    """
    return await request.read(), True


def _init_swagger_docs(app: web.Application, openapi_spec_file: Optional[str], project_slug: str,
                       routes: Iterable[Tuple[Callable, str, Callable]]) -> None:
    swagger = SwaggerDocs(app,
                          swagger_ui_settings=SwaggerUiSettings(path="/docs"),
                          info=SwaggerInfo(title=project_slug,
                                           version=version(project_slug),
                                           description='A HEA microservice'),
                          components=openapi_spec_file)
    swagger.register_media_type_handler('application/vnd.collection+json', application_json)
    swagger.register_media_type_handler('application/octet-stream', application_octet_stream)
    swagger.register_media_type_handler('text/plain', text_plain)
    swagger.add_routes([r[0](r[1], r[2]) for r in routes])






"""
Classes and functions for mocking Amazon S3 buckets.

This module assumes the moto package is installed. Do not import it into environments where moto will
not be available, for example, in any code that needs to run outside automated testing.
"""
from configparser import ConfigParser
from contextlib import AbstractContextManager
from typing import Optional, cast

from aiohttp import web
from heaobject.registry import Property
from heaobject.folder import AWSS3Folder
from heaobject.data import AWSS3FileObject
from heaobject.user import AWS_USER
from heaobject.source import AWS

from ..db.database import query_fixtures, Database

from ..db.aws import S3, S3Manager, create_client, S3ObjectPermissionContext
from ..db.mongo import Mongo
from ..oidcclaimhdrs import SUB
from heaobject.keychain import CredentialTypeVar, AWSCredentials
from heaobject.volume import AWSFileSystem, Volume, FileSystem
from heaobject.root import DesktopObjectDict, DesktopObject, PermissionContext, DesktopObjectTypeVar
from heaobject.account import AWSAccount
from heaobject.bucket import AWSBucket
from heaobject.awss3key import decode_key
from heaobject.user import NONE_USER
from heaobject.aws import S3Object
from aiohttp.web import Request

import boto3
from mypy_boto3_s3.literals import StorageClassType
from moto import mock_aws

from .util import freeze_time

from ..db.database import query_content, CollectionKey
from .mockmongo import MockMongo, MockMongoManager
from .mockdatabase import MockDatabase
from collections.abc import Sequence, AsyncIterator, Mapping, Callable
from io import BytesIO
import logging


class MockS3(S3, MockDatabase):
    """
    Overrides the S3 class' methods use moto instead of actual AWS. It mocks calls to get file system and credentials
    information.
    """

    async def get_file_system_and_credentials_from_volume(self, request: Request, volume_id: str) -> tuple[
        AWSFileSystem, CredentialTypeVar | None]:
        """
        Gets a file system and credentials using the provided volume id.

        :returns: a tuple with an AWSFileSystem instance and None.
        """
        return AWSFileSystem(), None

    async def get_credentials_from_volume(self, request: Request, volume_id: str) -> AWSCredentials | None:
        """
        Gets a credentials object using the provided volume id.

        :returns: None.
        """
        return AWSCredentials()

    async def get_account(self, request: Request, volume_id: str) -> AWSAccount | None:
        """
        Gets a mock AWS account.

        :return: a mock AWSAccount object.
        """
        return _get_account(request, volume_id)

    async def get_accounts(self, request: Request, volume_ids: Sequence[str]) -> AsyncIterator[tuple[AWSAccount, str]]:
        for volume_id in volume_ids:
            acct = await self.get_account(request, volume_id)
            if acct:
                yield acct, volume_id

    async def elevate_privileges(self, request: Request, credentials: AWSCredentials, lifespan: int | None = None) -> AWSCredentials:
        return AWSCredentials()


class MockS3WithMongo(S3, MockDatabase, Mongo):
    """
    Overrides the S3 class' methods use moto instead of actual AWS. It also stores volume and filesystem desktop
    objects in a mongo database. Moto is documented at https://github.com/spulec/moto.
    """

    async def generate_cloud_credentials(self, request: Request, arn: str, session_name: str,
                                         duration: int | None = None) -> AWSCredentials:
        """
        Create temporary credentials and return a newly created AWSCredentials object.

        :param request: the HTTP request (required).
        :param arn: The aws role arn that to be assumed (required).
        :param session_name: the session name to pass into AWS (required; typically is the user sub).
        :param duration: the time in seconds to assume the role. If None, it uses a default value, MIN_DURATION_SECONDS
        if the user is system|credentialsmanager, otherwise MAX_DURATION_SECONDS. The minimum is MIN_DURATION_SECONDS.
        :returns: an AWSCredentials object with the generated credentials.
        :raises ValueError: if an error occurs generating cloud credentials.
        """
        cred_dict_list: list[DesktopObjectDict] = []
        async for cred_dict in self.get_all(request, 'credentials'):
            cred_dict_list.append(cred_dict)
        cred = None
        if cred_dict_list and len(cred_dict_list) > 1:
            cred_dict = cred_dict_list[1]
            cred = AWSCredentials()
            cred.from_dict(cred_dict)
        elif cred_dict_list and len(cred_dict_list) == 1:
            cred_dict = cred_dict_list[0]
            cred = AWSCredentials()
            cred.from_dict(cred_dict)
        if cred is None:
            raise ValueError('No credentials')
        return cred

    async def get_account(self, request: Request, volume_id: str) -> AWSAccount | None:
        """
        Gets a mock AWS account.

        :param request: the HTTP request (required).
        :param volume_id: the volume id.
        :return: a mock AWSAccount object.
        """
        return _get_account(request, volume_id)


class MockS3WithMockMongo(S3, MockMongo):
    """
    Overrides the S3 class' methods use moto instead of actual AWS. It also stores volume and filesystem desktop
    objects in an in-memory data structure and mocks AWS microservices' attempts to access the file system and volumes
    microservices while requesting a boto3/moto client. Moto is documented at https://github.com/spulec/moto.
    """

    def __init__(self, config: Optional[ConfigParser], mongo: MockMongo | None = None, **kwargs):
        super().__init__(config=config, mongo=mongo, **kwargs)
        section = super().get_config_section()
        if config and section in config:
            database_section = config[section]
            self.__region_name: Optional[str] = database_section.get('RegionName', None)
            self.__aws_access_key_id: Optional[str] = database_section.get('AWSAccessKeyId', None)
            self.__aws_secret_access_key: Optional[str] = database_section.get('AWSSecretAccessKey', None)
            self.__expiration: Optional[str] = database_section.get('Expiration', None)
            self.__session_token: Optional[str] = database_section.get('SessionToken', None)
        else:
            self.__region_name = None
            self.__aws_access_key_id = None
            self.__aws_secret_access_key = None
            self.__expiration = None
            self.__session_token = None

    async def get_file_system_and_credentials_from_volume(self, request: Request, volume_id: str) -> tuple[
        AWSFileSystem, AWSCredentials | None]:
        """
        Gets an AWSFileSystem object and a credentials object either using the configuration passed into the
        constructor, or if not provided, the mock mongo in-memory database. If no credentials object is found, the
        AWSFileSystem and None are returned.

        :param request: the HTTP request (required).
        :param volume_id: the volume_id.
        :return: a tuple containing the file system and the credential object if found.
        """

        if self.__aws_secret_access_key is not None or self.__aws_access_key_id is not None:
            creds_ = AWSCredentials()
            creds_.where = self.__region_name
            creds_.account = self.__aws_access_key_id
            creds_.password = self.__aws_secret_access_key
            creds_.expiration = self.__expiration  # type:ignore[assignment]
            creds: AWSCredentials | None = creds_
        elif volume_id is None:
            creds = None
        else:
            creds = self._get_credential_by_volume(volume_id=volume_id)
        return AWSFileSystem(), creds

    async def get_credentials_from_volume(self, request: Request, volume_id: str | None) -> AWSCredentials | None:
        """
        Gets a credentials object either using the configuration passed into the constructor, or if not provided, the
        mock mongo in-memory database. If no credentials object is found, None is returned.

        :param request: the HTTP request (required).
        :param volume_id: the volume_id.
        :return: the credential object if found.
        """
        if self.__aws_secret_access_key is not None or self.__aws_access_key_id is not None:
            creds_ = AWSCredentials()
            creds_.where = self.__region_name
            creds_.account = self.__aws_access_key_id
            creds_.password = self.__aws_secret_access_key
            creds_.expiration = self.__expiration  # type:ignore[assignment]
            creds: AWSCredentials | None = creds_
        elif volume_id is None:
            creds = None
        else:
            creds = self._get_credential_by_volume(volume_id=volume_id)
        return creds

    async def get_volumes(self, request: Request, file_system_type_or_type_name: str | type[FileSystem],
                          account_ids: Sequence[str] | None = None) -> AsyncIterator[Volume]:
        """
        An asynchronous iterator of the volumes to which the user has access, retrieved from the mock mongo in-memory
        database.

        :param request: the HTTP request (required).
        :param file_system_type_or_type_name: filter by a file system type or type name (required).
        :return: an asynchronous iterator of Volume objects.
        """
        if isinstance(file_system_type_or_type_name, type) and issubclass(file_system_type_or_type_name, DesktopObject):
            file_system_type_name_ = file_system_type_or_type_name.get_type_name()
        else:
            file_system_type_name_ = str(file_system_type_or_type_name)
        for volume_dict in self.get_desktop_objects_by_collection('volumes'):
            if volume_dict.get('file_system_type', AWSFileSystem.get_type_name()) == file_system_type_name_:
                volume = Volume()
                volume.from_dict(volume_dict)
                if not account_ids or volume.account_id in account_ids:
                    yield volume

    async def get_property(self, app: web.Application, name: str) -> Property | None:
        """
        Gets the value of the requested property from the mock mongo in-memory database.

        :param app: the aiohttp web.Application object (required).
        :param name: the name of the property (required).
        :return: a Property object or None if not found.
        """
        prop_dict_list = [prop_dict for prop_dict in self.get_desktop_objects_by_collection('properties')
                          if prop_dict.get('name', None) == name]
        prop = None
        if prop_dict_list and len(prop_dict_list) > 0:
            prop_dict = prop_dict_list[0]
            prop = Property()
            prop.from_dict(prop_dict)
        return prop if prop else None

    async def update_credentials(self, request: Request, credentials: AWSCredentials) -> None:
        """
        This method is for updating the provided credentials object in the database, but this implementation does
        nothing.

        :param request: the HTTP request (required).
        :param credentials: the AWSCredentials to update (required).
        """
        pass

    async def get_account(self, request: Request, volume_id: str) -> AWSAccount | None:
        """
        Gets a mock AWS account.

        :return: a mock AWSAccount object.
        """
        return _get_account(request, volume_id)

    async def elevate_privileges(self, request: Request, credentials: AWSCredentials, lifespan: int | None = None) -> AWSCredentials:
        return AWSCredentials()

    def _get_credential_by_volume(self, volume_id: str) -> Optional[AWSCredentials]:
        volume_dict = self.get_desktop_object_by_collection_and_id('volumes', volume_id)
        if volume_dict is None:
            raise ValueError(f'No volume found with id {volume_id}')
        volume = Volume()
        volume.from_dict(volume_dict)
        if volume.credential_id is None:
            creds = None
        else:
            credentials_dict = self.get_desktop_object_by_collection_and_id('credentials', volume.credential_id)
            if credentials_dict is None:
                raise ValueError(f'No credentials with id {volume.credential_id}')
            creds = AWSCredentials()
            creds.from_dict(credentials_dict)
        return creds


class MockS3Manager(S3Manager):
    """
    Database manager for mocking AWS S3 buckets with moto. Mark test fixture data that is managed in S3 buckets with
    this database manager in testing environments. Furthermore, connections to boto3/moto clients normally require
    access to the registry and volume microservices. This database manager does not mock those connections, and actual
    registry and volume microservices need to be running, as is typical in integration testing environments. For unit
    testing, see MockS3ManagerWithMockMongo, which also mocks the connections to the registry and volume microservices.
    """

    @classmethod
    def get_environment_updates(cls) -> dict[str, str]:
        result = super().get_environment_updates()
        result.update({'AWS_ACCESS_KEY_ID': 'testing',
                       'AWS_SECRET_ACCESS_KEY': 'testing',
                       'AWS_SECURITY_TOKEN': 'testing',
                       'AWS_SESSION_TOKEN': 'testing',
                       'AWS_DEFAULT_REGION': 'us-east-1'
                       })
        return result

    @classmethod
    def get_context(cls) -> list[AbstractContextManager]:
        result = super().get_context()
        result.extend([mock_aws(), freeze_time()])
        return result

    def get_database(self) -> S3:
        return MockS3(self.config)

    def insert_desktop_objects(self, desktop_objects: Optional[Mapping[CollectionKey, Sequence[DesktopObjectDict]]]):
        super().insert_desktop_objects(desktop_objects)
        logger = logging.getLogger(__name__)

        for coll, objs in query_fixtures(desktop_objects, db_manager=self).items():
            logger.debug('Inserting %s collection object %s', coll, objs)
            inserters = self.get_desktop_object_inserters()
            if coll in inserters:
                inserters[coll](objs)

    def insert_content(self, content: Optional[Mapping[CollectionKey, Mapping[str, bytes]]]):
        super().insert_content(content)
        if content is not None:
            client = create_client('s3')
            try:
                for key, contents in query_content(content, db_manager=self).items():
                    if key == 'awss3files':
                        for id_, content_ in contents.items():
                            bucket_, actual_content = content_.split(b'|', 1)
                            bucket = bucket_.decode('utf-8')
                            with BytesIO(actual_content) as f:
                                client.upload_fileobj(Fileobj=f, Bucket=bucket, Key=decode_key(id_))
                    else:
                        raise KeyError(f'Unexpected key {key}')
            finally:
                client.close()

    @classmethod
    def get_desktop_object_inserters(cls) -> dict[str, Callable[[list[DesktopObjectDict]], None]]:
        return {'buckets': cls.__bucket_inserter,
                'awss3folders': cls.__awss3folder_inserter,
                'awss3files': cls.__awss3file_inserter}

    @classmethod
    def __organization_inserter(cls):
        cls.__create_organization()

    @classmethod
    def __awsaccount_inserter(cls):
        pass  # moto automatically creates one account to use. Creating your own accounts doesn't seem to work.


    @classmethod
    def __awss3file_inserter(cls, v):
        for awss3file_dict in v:
            awss3file = AWSS3FileObject()
            awss3file.from_dict(awss3file_dict)
            cls.__create_awss3file(awss3file)

    @classmethod
    def __awss3folder_inserter(cls, v):
        for awss3folder_dict in v:
            awss3folder = AWSS3Folder()
            awss3folder.from_dict(awss3folder_dict)
            cls.__create_awss3folder(awss3folder)

    @classmethod
    def __bucket_inserter(cls, v):
        for bucket_dict in v:
            awsbucket = AWSBucket()
            awsbucket.from_dict(bucket_dict)
            cls.__create_bucket(awsbucket)

    @staticmethod
    def __create_organization():
        client = boto3.client('organizations')
        try:
            client.create_organization(FeatureSet='ALL')
        finally:
            client.close()

    @staticmethod
    def __create_bucket(bucket: AWSBucket):
        client = create_client('s3')
        try:
            if bucket is not None:
                if bucket.name is None:
                    raise ValueError('bucket.name cannot be None')
                else:
                    if bucket.region != 'us-east-1' and bucket.region:
                        client.create_bucket(Bucket=bucket.name,
                                            CreateBucketConfiguration={'LocationConstraint': bucket.region})
                    else:
                        client.create_bucket(Bucket=bucket.name)
                    client.put_bucket_versioning(Bucket=bucket.name,
                                                VersioningConfiguration={'MFADelete': 'Disabled', 'Status': 'Enabled'})
        finally:
            client.close()

    @staticmethod
    def __create_awss3folder(awss3folder: AWSS3Folder):
        assert awss3folder.bucket_id is not None, 'awss3file must have a non-None bucket_id attribute'
        assert awss3folder.key is not None, 'awss3file must have a non-None key attribute'
        client = create_client('s3')
        try:
            client.put_object(Bucket=awss3folder.bucket_id, Key=awss3folder.key, StorageClass='STANDARD')
        finally:
            client.close()

    @staticmethod
    def __create_awss3file(awss3file: AWSS3FileObject):
        assert awss3file.bucket_id is not None, 'awss3file must have a non-None bucket_id attribute'
        assert awss3file.key is not None, 'awss3file must have a non-None key attribute'
        assert awss3file.storage_class is not None and awss3file.storage_class.name is not None, 'awss3file must have a non-None storage_class with a non-None name attribute'
        client = create_client('s3')
        try:
            storage_class_name = cast(StorageClassType, awss3file.storage_class.name)
            client.put_object(Bucket=awss3file.bucket_id, Key=awss3file.key, StorageClass=storage_class_name)
        finally:
            client.close()



class MockS3WithMockMongoManager(MockS3Manager, MockMongoManager):
    """
    Database manager for mocking AWS S3 buckets with moto. Mark test fixture data that is managed in S3 buckets with
    this database manager in unit test environments. Furthermore, connections to boto3/moto clients normally require
    access to the registry and volume microservices. This database manager mocks those connections. Mark
    component, volume, and filesystem test collections with this database manager to make them available in unit
    testing environments. This class is not designed to be subclassed.
    """

    def get_database(self) -> MockS3WithMockMongo:
        return MockS3WithMockMongo(config=self.config, mongo=self.get_mongo())

    @classmethod
    def database_types(self) -> list[str]:
        return ['system|awss3', 'system|mongo']


def _get_account(request: Request, volume_id: str) -> AWSAccount | None:
    """
    Gets a mock AWS account.

    :return: a mock AWSAccount object.
    """
    account_id = '123456789012'
    a = AWSAccount()
    a.id = account_id
    a.display_name = account_id
    a.name = 'master'
    a.owner = AWS_USER
    a.source = AWS
    a.source_detail = AWS
    a.email_address = 'master@example.com'

    return a



"""
HEA Server Framework is a framework library for creating HEA microservices.

Types of microservices

The registry microservice manages a listing of all microservices that the current instance of HEA knows about.

Trash microservices serve items that have been marked for permanent deletion but have not been permanently deleted yet.
The registry service may have at most one trash microservice registered for a given desktop object type, file system
type, and file system name combination. Desktop object types with no registered trash microservice are assumed not to
have a trash and are deleted permanently.

Class in this package and all microservices have the following conventions for object attributes:
* Private attributes' names are prefixed with a double underscore.
* Protected attributes' names are prefixed with a single underscore. "Protected" is defined as accessible only to
the class in which it's defined and subclasses. Python does not enforce protected access, but uses of protected
attributes outside of subclasses may break even in patch releases.

HEA adheres to the following URL conventions:
* access_token is a reserved query parameter. It is an alternative to passing the Bearer token in the request's
Authorization header.
* get requests may accept a data parameter with values true/false/yes/no/t/f/y/n (case insensitive) that controls
whether the desktop object and the requesting user's permissions for it are included in the response. Endpoints may
ignore this parameter.
* Getting a desktop object's contents: append /content to the object's URL. Optionally pass the mode query parameter,
which has two possible values, download (to download the content from a browser) or open (to open the content in a
browser). The default is download.

HEA objects support various operations similar to those in a file system, such as copy, move, delete, and rename. These
operations are not guaranteed to be atomic, and may not be supported by all objects. Here are some expected behaviors:
* Objects may be versioned or unversioned, indicated by the presence of a version attribute. Versioned objects have a
  version history, while unversioned objects do not. Versioned objects support permanently deleting old versions. They
  also support making a selected old version the current version, which moves the selected version to the top of the
  version history, potentially changing the version's version id in the process.
* Copying an object to a different location creates a new object with the same content as the original. For versioned
  objects, only the latest version is copied. Copying an object to the same location must not change the object.
* Moving an object to a different location may change object metadata without physically moving the object, or it may
  create a new object with the same content as the original and deletes the original object. For versioned objects,
  moving an object within a volume also moves all versions of the original object to the new location. Moving an object
  to a different volume only moves the latest version of the object and discards all versions in the original location.
  For volumes that support the Trash, the original object and its versions may be restorable from the Trash.
* Deleting an object may either permanently delete it or move it to the Trash, depending on whether the object's volume
  supports the Trash. All the object's versions must also be moved or permanently deleted. A "move" to the Trash may be
  implemented as an actual move, or the object may stay in the same location, including all its versions for versioned
  objects, but be annotated as deleted. The user must be warned if the object is about to be permanently deleted.
* Renaming an object changes its name, and depending on the underlying file system, may be implemented as a metadata
  change or an actual move. For versioned objects, all versions of the object are renamed.

REST API endpoints that return multiple resources use an undefined order for the returned resources by default. To
specify an order, use the sort query parameter with a value of either asc or desc (ascending or descending,
respectively). The parameter's values are case insensitive. A sort query parameter will apply to an endpoint-specific
default attribute. An endpoint may allow overriding the default attribute with a sort_attr query parameter. To specify
different sort orders for different attributes, an endpoint may support specifying multiple sort query parameters
paired with multiple sort_attr query parameters. For example, `?sort=asc&sort_attr=name&sort=desc&sort_attr=created_at`
sorts the results by name in ascending order and by created_at in descending order. The sort_attr query parameter
values are case-sensitive.

As mentioned in the heaobject documentation, a folder with id "root" denotes the root of a volume or volume-like
object, like an AWS S3 bucket. It is a view of the volume-like object. The root must never be returned by any folder
endpoint. However, there may be requests for the root folder's items or requests for a form template to create or
update objects within the root folder. In the former case, the folder's items can be manipulated by an endpoint. In the
latter case, when getting the form template, return the volume-like object and use the volume-like object's attributes
to populate the form template.
"""
import logging
logging.getLogger(__name__).addHandler(logging.NullHandler())



from unittest import IsolatedAsyncioTestCase, TestCase

from aiohttp.test_utils import make_mocked_request
from heaobject.aws import S3StorageClass, S3ArchiveDetailState
from heaobject.data import AWSS3FileObject
from heaobject.group import SUPERADMIN_GROUP
from heaobject.root import Permission
from moto import mock_aws
from heaserver.service.db import awsservicelib
from botocore.exceptions import ClientError

import heaserver.service.db.aws
from heaserver.service.db.aws import S3ObjectPermissionContext
from heaserver.service.testcase.mockaws import MockS3


class TestAWSServiceLib(TestCase):

    def test_handle_client_error_404(self):
        c = ClientError(error_response={'Error': {'Code': heaserver.service.db.aws.CLIENT_ERROR_404}}, operation_name='foo')
        self.assertEqual(404, awsservicelib.handle_client_error(c).status)

    def test_handle_client_error_no_such_bucket(self):
        c = ClientError(error_response={'Error': {'Code': heaserver.service.db.aws.CLIENT_ERROR_NO_SUCH_BUCKET}}, operation_name='foo')
        self.assertEqual(404, awsservicelib.handle_client_error(c).status)

    def test_handle_client_error_unknown(self):
        c = ClientError(error_response={'Error': {'Code': "It's wrecked"}},
                        operation_name='foo')
        self.assertEqual(500, awsservicelib.handle_client_error(c).status)


class TestAsyncAWSServiceLib(IsolatedAsyncioTestCase):
    @mock_aws()
    def run(self, result):
        super().run(result)

    def setUp(self):
        self.request = make_mocked_request('GET', '/')
        self.request.app['HEA_db'] = MockS3(None)
        self.perm_context: S3ObjectPermissionContext = S3ObjectPermissionContext(self.request)
        from types import MethodType
        async def mock_is_super_admin(self) -> bool:
            return False
        self.perm_context.is_super_admin = MethodType(mock_is_super_admin, self.perm_context)

    async def test_handle_display_name_attribute_permissions_glacier_deep_archive(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.storage_class = S3StorageClass.DEEP_ARCHIVE
        obj.archive_detail_state = S3ArchiveDetailState.ARCHIVED
        obj.key = 'test_key'
        self.assertEqual([Permission.VIEWER], await self.perm_context.get_attribute_permissions(obj, 'display_name'))

    async def test_handle_display_name_attribute_permissions_glacier_instant_retrieval(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.storage_class = S3StorageClass.GLACIER_IR
        obj.archive_detail_state = S3ArchiveDetailState.NOT_ARCHIVED
        obj.key = 'test_key'
        self.assertEqual(set([Permission.VIEWER, Permission.EDITOR]), set(await self.perm_context.get_attribute_permissions(obj, 'display_name')))

    async def test_handle_display_name_attribute_permissions_standard(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.storage_class = S3StorageClass.STANDARD
        obj.key = 'test_key'
        self.assertEqual(set([Permission.VIEWER, Permission.EDITOR]), set(await self.perm_context.get_attribute_permissions(obj, 'display_name')))

    async def test_handle_display_name_attribute_permissions_unknown(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.key = 'test_key'
        self.assertEqual(set([Permission.VIEWER, Permission.EDITOR]), set(await self.perm_context.get_attribute_permissions(obj, 'display_name')))

    async def test_handle_display_name_attribute_permissions_restoring(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.key = 'test_key'
        obj.storage_class = S3StorageClass.DEEP_ARCHIVE
        obj.archive_detail_state = S3ArchiveDetailState.RESTORING
        self.assertEqual([Permission.VIEWER], await self.perm_context.get_attribute_permissions(obj, 'display_name'))

    async def test_handle_display_name_attribute_permissions_restored(self):
        obj: AWSS3FileObject = AWSS3FileObject()
        obj.key = 'test_key'
        obj.storage_class = S3StorageClass.DEEP_ARCHIVE
        obj.archive_detail_state = S3ArchiveDetailState.RESTORED
        self.assertEqual(set([Permission.VIEWER, Permission.EDITOR]), set(await self.perm_context.get_attribute_permissions(obj, 'display_name')))







from .componenttestcase import ComponentTestCase
from heaserver.service.testcase.mixin import PutMixin


class TestPutComponent(ComponentTestCase, PutMixin):  # type: ignore

    async def test_put_status_invalid_base_url(self):
        await self._test_invalid({'base_url': 2})

    async def test_put_status_invalid_resource(self):
        await self._test_invalid({'resources': [2]})



